{% extends "base.html" %}

{% block extra_scripts %}
   {% block mellon_extra_scripts %}
   {% endblock %}
{% endblock %}

{% block content %}
   {% block mellon_content %}
   {% endblock %}
{% endblock %}
